import glob
import os
import posixpath
import shutil
import sys
import tempfile
from pathlib import Path
from tempfile import NamedTemporaryFile

import pytest
from bencodepy.exceptions import BencodeDecodeError

import qbt_migrate
from qbt_migrate.classes import FastResume, QBTBatchMove


@pytest.fixture
def linux_env(monkeypatch):
    monkeypatch.setenv("HOME", "/home/test")
    monkeypatch.setattr(sys, "platform", "linux")
    monkeypatch.setattr(os, "path", posixpath)


@pytest.fixture
def temp_dir():
    temp_dir = Path(tempfile.mkdtemp())  # Get a tmp dir
    yield temp_dir
    shutil.rmtree(temp_dir)


@pytest.fixture
def temp_file(mode="w"):
    with NamedTemporaryFile(mode, delete=False) as f:
        yield f.name, f


@pytest.mark.parametrize(
    "bt_backup_path, expected_bt_backup_path",
    [(None, "/home/test/.local/share/data/qBittorrent/BT_backup"), ("/some/path", "/some/path")],
)
def test_qbt_batch_move_init(linux_env, bt_backup_path, expected_bt_backup_path):
    qbt = QBTBatchMove(bt_backup_path)
    assert qbt.bt_backup_path == expected_bt_backup_path
    assert isinstance(qbt.discovered_files, set)


def test_qbt_batch_move_discover_relevant_fast_resume(temp_dir):
    fast_resume_files = list(QBTBatchMove.discover_relevant_fast_resume(temp_dir, "", False, True))
    assert fast_resume_files == []

    for file in glob.glob("./tests/test_files/good*.fastresume"):
        file = Path(file)
        print(f"Copying {file} to {temp_dir / file.name}")
        shutil.copy(file, temp_dir / file.name)
    fast_resume_files = list(QBTBatchMove.discover_relevant_fast_resume(temp_dir, "", False, True))
    assert len(fast_resume_files) == 2
    fast_resume_files = list(QBTBatchMove.discover_relevant_fast_resume(temp_dir, "/some/test", False, True))
    assert len(fast_resume_files) == 1
    assert fast_resume_files[0].save_path.startswith("/some/test") is True
    fast_resume_files = list(QBTBatchMove.discover_relevant_fast_resume(temp_dir, r"/some/(\w+)/.*$", True, True))
    assert len(fast_resume_files) == 1
    assert fast_resume_files[0].save_path.startswith("/some/test") is True

    for file in glob.glob("./tests/test_files/bad*.fastresume"):
        file = Path(file)
        print(f"Copying {file} to {temp_dir / file.name}")
        shutil.copy(file, temp_dir / file.name)
    with pytest.raises(BencodeDecodeError):
        list(QBTBatchMove.discover_relevant_fast_resume(temp_dir, "/some/test", False, True))

    fast_resume_files = list(QBTBatchMove.discover_relevant_fast_resume(temp_dir, "/some/test", False, False))
    assert len(fast_resume_files) == 1


def test_qbt_batch_move_run_not_a_dir(temp_file):
    # Test path not exists
    qbt = QBTBatchMove("/not/a/valid/path")
    pytest.raises(NotADirectoryError, qbt.run, "existing_path", "new_path")

    # Test path exists, but not a directory
    qbt = QBTBatchMove(temp_file[0])
    pytest.raises(NotADirectoryError, qbt.run, "existing_path", "new_path")


def test_qbt_batch_move_run_create_backup_and_backup_folder(monkeypatch, temp_dir):
    class MockCaller:
        def __init__(self):
            self.called = False

        def mock_create_backup(self, directory: str, archive: str):
            self.called = True
            assert directory == temp_dir
            assert archive.startswith(os.path.join(temp_dir, "fastresume_backup"))
            assert archive.endswith(".zip")

    mock = MockCaller()
    monkeypatch.setattr(qbt_migrate.classes, "backup_folder", mock.mock_create_backup)

    qbt = QBTBatchMove(temp_dir)
    qbt.run("existing_path", "new_path", create_backup=False)
    assert mock.called is False  # Test `backup_folder` is not called
    qbt.run("existing_path", "new_path", create_backup=True)
    assert mock.called is True  # Test `backup_folder` is called

    mock.called = False  # Reset MockCaller
    qbt = QBTBatchMove(temp_dir)
    qbt.backup_folder(temp_dir, os.path.join(temp_dir, "fastresume_backup.zip"))
    assert mock.called is True


def test_qbt_batch_move_run_replace_paths(monkeypatch, temp_dir):
    class MockFastResume(FastResume):
        def __init__(self, *_, **__):
            self.called = False
            self._data = {"save_path": "existing_path", "qBt-savePath": "existing_path"}

        def replace_paths(self, *_, **__):
            self.called = True

    with open(temp_dir / "test.fastresume", "w") as _:
        pass

    monkeypatch.setattr(qbt_migrate.classes, "FastResume", MockFastResume)

    qbt = QBTBatchMove(temp_dir)
    qbt.run("existing", "new", create_backup=False)
    assert len(qbt.discovered_files) == 1
    for fast_resume in qbt.discovered_files:
        assert fast_resume.called


def test_qbt_batch_move_update_fastresume(monkeypatch):
    class MockFastResume(FastResume):
        def __init__(self):
            self.called = False

        def replace_paths(self, *_, **__):
            self.called = True

    mock_fast_resume = MockFastResume()
    QBTBatchMove.update_fastresume(mock_fast_resume, "existing_path", "new_path")
    assert mock_fast_resume.called

    pytest.raises(
        TypeError, qbt_migrate.classes.QBTBatchMove.update_fastresume, "not_fast_resume_object", "existing", "new"
    )


def test_fastresume_init(temp_dir):
    for file in glob.glob("./tests/test_files/good*.fastresume"):
        file = Path(file)
        print(f"Copying {file} to {temp_dir / file.name}")
        shutil.copy(file, temp_dir / file.name)
        fast_resume = FastResume(temp_dir / file.name)
        assert fast_resume.save_path is not None

    with pytest.raises(FileNotFoundError):
        FastResume("/not/a/valid/file")


def test_fastresume_properties(temp_dir):
    for file in glob.glob("./tests/test_files/good*.fastresume"):
        file = Path(file)
        print(f"Copying {file} to {temp_dir / file.name}")
        shutil.copy(file, temp_dir / file.name)
        fast_resume = FastResume(temp_dir / file.name)
        assert fast_resume.file_path == str(temp_dir / file.name)
        assert fast_resume.backup_filename.startswith(str(temp_dir / file.name))
        assert fast_resume.backup_filename.endswith(".bkup")
        assert fast_resume.save_path == fast_resume._data["save_path"]
        assert fast_resume.qbt_save_path == fast_resume._data["qBt-savePath"]
        assert (
            fast_resume.mapped_files == fast_resume._data["mapped_files"]
            if "mapped_files" in fast_resume._data
            else fast_resume.mapped_files is None
        )


class MockFastResumeSaveCaller:
    """
    Simple mock caller for `FastResume.save` method that provides minimal tracking
    """

    def __init__(self):
        self.called = False
        self.saved_file_names = set()

    def mock_save(self, fn=None):
        self.called = True
        if fn is not None:
            self.saved_file_names.add(fn)

    def mock_write(self, _, fn):
        self.mock_save(fn)

    def reset(self):
        self.__init__()


def test_fastresume_set_save_path(monkeypatch):
    fast_resume = FastResume("./tests/test_files/good.fastresume")
    mock = MockFastResumeSaveCaller()
    monkeypatch.setattr(fast_resume, "save", mock.mock_save)

    # Test invalid save key
    with pytest.raises(KeyError):
        fast_resume.set_save_path("/a/path", "not-a-valid-key")
    # Explicitly test setting of key
    fast_resume.set_save_path("/this/is/a/test", save_file=False, create_backup=False)
    fast_resume.set_save_path("/this/is/a/test", key="qBt-savePath", save_file=False, create_backup=False)
    assert fast_resume.save_path == "/this/is/a/test"
    assert fast_resume.qbt_save_path == "/this/is/a/test"

    mock.reset()
    # Test target_os None
    fast_resume.set_save_path("/target\\is/none", save_file=False, create_backup=False)
    assert fast_resume.save_path == "/target\\is/none"
    # Test target_os windows
    fast_resume.set_save_path("C:/target/windows", target_os="windows", save_file=False, create_backup=False)
    assert fast_resume.save_path == "C:\\target\\windows"
    # Test target_os linux
    fast_resume.set_save_path("\\target\\linux", target_os="linux", save_file=False, create_backup=False)
    assert fast_resume.save_path == "/target/linux"
    # Test target_os mac
    fast_resume.set_save_path("\\target\\mac", target_os="mac", save_file=False, create_backup=False)
    assert fast_resume.save_path == "/target/mac"

    mock.reset()
    # Test save_file
    fast_resume.set_save_path("/this/is/a/test", save_file=True, create_backup=False)
    assert mock.called is True
    assert len(mock.saved_file_names) == 0

    mock.reset()
    # Test create_backup=False
    fast_resume.set_save_path("test", save_file=False, create_backup=False)
    assert mock.called is False
    # Test create_backup=True
    fast_resume.set_save_path("test", save_file=False, create_backup=True)
    assert mock.called
    assert len(mock.saved_file_names) == 1
    for file_name in mock.saved_file_names:
        assert "good.fastresume." in file_name
        assert file_name.endswith(".bkup")


def test_fastresume_set_save_paths(monkeypatch):
    fast_resume = FastResume("./tests/test_files/good.fastresume")
    mock = MockFastResumeSaveCaller()
    monkeypatch.setattr(fast_resume, "save", mock.mock_save)

    # Test empty path ValueError
    with pytest.raises(ValueError):
        fast_resume.set_save_paths("", create_backup=False, save_file=False)
    # Test default call
    fast_resume.set_save_paths("/this/is/a/path")
    assert fast_resume.save_path == "/this/is/a/path"
    assert fast_resume.qbt_save_path == "/this/is/a/path"
    # Test with differing qBt-savePath
    fast_resume.set_save_paths("/this/is/a/path", "/this/is/a/qbt/path")
    assert fast_resume.save_path == "/this/is/a/path"
    assert fast_resume.qbt_save_path == "/this/is/a/qbt/path"

    # Test mapped_files convert slashes (replacing of paths happens with `replace_paths`)
    fast_resume.set_save_paths("C:/this/is/a/path", target_os="windows")
    assert len(fast_resume.mapped_files) > 0
    for mapped_file in fast_resume.mapped_files:
        assert mapped_file.startswith("\\some\\test\\path\\")

    # Test missing qBt-savePath key
    del fast_resume._data["qBt-savePath"]
    fast_resume.set_save_paths("/this/is/a/path")
    assert fast_resume.save_path == "/this/is/a/path"
    assert fast_resume.qbt_save_path == "/this/is/a/path"


def test_fastresume_save(monkeypatch):
    fast_resume = FastResume("./tests/test_files/good.fastresume")
    mock = MockFastResumeSaveCaller()
    monkeypatch.setattr(qbt_migrate.classes.bencode, "write", mock.mock_write)

    # Test default call
    fast_resume.save()
    assert mock.called
    assert len(mock.saved_file_names) == 1
    for file_name in mock.saved_file_names:
        assert file_name.endswith("good.fastresume")

    mock.reset()
    # Test calling with file name
    fast_resume.save("test_file")
    assert mock.called
    assert len(mock.saved_file_names) == 1
    for file_name in mock.saved_file_names:
        assert file_name == "test_file"


def test_fastresume_replace_paths(monkeypatch):
    fast_resume = FastResume("./tests/test_files/good.fastresume")
    fast_resume.replace_paths("/some/test", "/a/new/test", save_file=False, create_backup=False)
    assert fast_resume.save_path == "/a/new/test/path"
    assert fast_resume.qbt_save_path == "/a/new/test/path"
    assert len(fast_resume.mapped_files) > 0
    for mapped_file in fast_resume.mapped_files:
        assert mapped_file.startswith("/a/new/test/path")


def test_fastresume_replace_paths_regex(monkeypatch):
    fast_resume = FastResume("./tests/test_files/good.fastresume")
    fast_resume.replace_paths(r"/some/(\w+)/.*$", r"/\1/regex", True, save_file=False, create_backup=False)
    assert fast_resume.save_path == "/test/regex"
    assert fast_resume.qbt_save_path == "/test/regex"
    assert len(fast_resume.mapped_files) > 0
    for mapped_file in fast_resume.mapped_files:
        assert mapped_file.startswith("/test/regex")



import ntpath
import os
import posixpath
import re
import sys
import tempfile
import zipfile
from pathlib import Path

import pytest

from qbt_migrate.methods import backup_folder, convert_slashes, discover_bt_backup_path


def test_backup_folder():
    temp_dir = Path(tempfile.mkdtemp())  # Get a tmp dir
    fastresume_file_count = 20  # Number of test fastresume files to build
    torrent_file_count = 20  # Number of test torrent files to build
    invalid_file_count = 20  # Number of invalid files to build

    # Build test files
    for x in range(fastresume_file_count):
        with open(temp_dir / f"test_{x}.fastresume", "w") as f:
            f.write(f"This is file {x}")

    for x in range(torrent_file_count):
        with open(temp_dir / f"test_{x}.torrent", "w") as f:
            f.write(f"This is file {x}")

    for x in range(invalid_file_count):
        with open(temp_dir / f"test_{x}.invalid", "w") as f:
            f.write(f"This is invalid file {x}")

    # 'test.zip' will fail the filename check in archive if it gets archived
    archive_path = temp_dir / "test.zip"
    backup_folder(temp_dir, archive_path)  # Call backup_folder with torrent files included

    # Validate zip file
    with zipfile.ZipFile(archive_path, "r") as archive:
        assert archive.testzip() is None  # Check no errors
        # Assert that it has the correct file count
        assert len(archive.infolist()) == fastresume_file_count + torrent_file_count
        for file in archive.infolist():
            # Get the integer of the file
            filename = re.findall(r"test_(\d+).(?:fastresume)?(?:torrent)?", file.filename)
            assert len(filename) != 0  # Ensure we got the integer
            filename = filename[0]
            assert archive.read(file.filename) == bytes(f"This is file {filename}", "utf-8")  # Integrity Check

    archive_path = temp_dir / "test2.zip"
    backup_folder(temp_dir, archive_path, False)

    # Validate zip file
    with zipfile.ZipFile(archive_path, "r") as archive:
        assert archive.testzip() is None  # Check no errors
        # Assert that it has the correct file count
        assert len(archive.infolist()) == fastresume_file_count
        for file in archive.infolist():
            # Get the integer of the file
            filename = re.findall(r"test_(\d+).fastresume", file.filename)
            assert len(filename) != 0  # Ensure we got the integer
            filename = filename[0]
            assert archive.read(file.filename) == bytes(f"This is file {filename}", "utf-8")  # Integrity Check


@pytest.mark.parametrize(
    "path, target_os, expected_path",
    [
        ("C:/some/path", "windows", "C:\\some\\path"),
        ("C:\\some\\path", "windows", "C:\\some\\path"),
        ("\\some\\path\\for\\linux", "linux", "/some/path/for/linux"),
        ("/some/path/for/linux", "linux", "/some/path/for/linux"),
        ("\\some\\path\\for\\mac", "mac", "/some/path/for/mac"),
        ("/some/path/for/mac", "mac", "/some/path/for/mac"),
        (None, "invalid_os", None),
    ],
)
def test_convert_slashes(path: str, target_os: str, expected_path: str):
    if target_os == "invalid_os":
        with pytest.raises(ValueError):
            convert_slashes(path, target_os)
    else:
        assert convert_slashes(path, target_os) == expected_path


@pytest.mark.parametrize(
    "system, expected_path, envvar_overrides",
    [
        (
            ("win32", ntpath),
            "C:\\Users\\test\\AppData\\Local\\qBittorrent\\BT_backup",
            {"localappdata": "C:\\Users\\test\\AppData\\Local"},
        ),
        (("linux", posixpath), "/home/test/.local/share/data/qBittorrent/BT_backup", {"HOME": "/home/test"}),
        (("docker", posixpath), "/config/qBittorrent/BT_backup", {}),
    ],
)
def test_discover_bt_backup_path(monkeypatch, system: tuple, expected_path: str, envvar_overrides: dict):
    for env, val in envvar_overrides.items():
        monkeypatch.setenv(env, val)
    system, os_path_lib = system
    monkeypatch.setattr(os, "path", os_path_lib)
    if system == "docker":
        system = "linux"

        def mock_docker_env(_):
            return True

        monkeypatch.setattr(os.path, "isfile", mock_docker_env)
        monkeypatch.setattr(os.path, "isdir", mock_docker_env)

    monkeypatch.setattr(sys, "platform", system)
    assert discover_bt_backup_path() == expected_path



import argparse
import logging
import sys

from . import QBTBatchMove, discover_bt_backup_path


logger = logging.getLogger(__name__)


def parse_args(args=None):
    parser = argparse.ArgumentParser()
    parser.add_argument("-e", "--existing-path", help="Existing root of path to look for.")
    parser.add_argument("-n", "--new-path", help="New root path to replace existing root path with.")
    parser.add_argument(
        "-r",
        "--regex",
        help="Existing and New paths are regex patterns with capture groups.",
        action="store_true",
        default=None,
    )
    parser.add_argument(
        "-t",
        "--target-os",
        help="Target OS (converts slashes). "
        "Default will auto-detect if conversion is needed "
        "based on existing vs new.",
        choices=["Windows", "Linux", "Mac"],
    )
    parser.add_argument(
        "-b", "--bt-backup-path", help="BT_backup Path Override. " f"Default is {discover_bt_backup_path()}"
    )
    parser.add_argument(
        "-s",
        "--skip-bad-files",
        help="Skips bad .fastresume files instead of exiting. " "Default behavior is to exit.",
        action="store_true",
        default=False,
    )

    parser.add_argument(
        "-l", "--log-level", help="Log Level, Default is INFO.", choices=["DEBUG", "INFO"], default="INFO"
    )

    return parser.parse_args(args)


def main():
    args = parse_args(sys.argv[1:])
    logging.basicConfig()
    logger.setLevel(args.log_level)
    logging.getLogger("qbt_migrate").setLevel(args.log_level)
    logging.getLogger("qbt_migrate").propagate = True
    qbm = QBTBatchMove()
    if args.bt_backup_path is not None:
        qbm.bt_backup_path = args.bt_backup_path
    else:
        bt_backup_path = input(f"BT_backup Path {qbm.bt_backup_path}: ")
        if bt_backup_path.strip():
            qbm.bt_backup_path = bt_backup_path
    if args.existing_path is None:
        args.existing_path = input("Existing Path: ")
    if args.new_path is None:
        args.new_path = input("New Path: ")

    # Get Valid Regex Input
    if args.regex is None:
        while (answer := input("Regex Paths with Capture Groups [y/N]: ").lower().strip()) not in (
            "y",
            "yes",
            "n",
            "no",
            "",
        ):
            print("Please answer y, n, yes, or no")
        args.regex = answer.lower().strip() in ["y", "yes"]

    # Get Valid Target OS Input
    if args.target_os is None:
        while (answer := input("Target OS (Windows, Linux, Mac, Blank for auto-detect): ").lower().strip()) not in (
            "windows",
            "linux",
            "mac",
            "",
        ):
            print("Please answer Windows, Linux, or Mac")
        args.target_os = answer.lower().strip()

    # Handle Target OS Auto-Detect if not specified
    if not args.target_os.strip():
        if "/" in args.existing_path and "\\" in args.new_path:
            logger.info("Auto detected target OS change. Will convert slashes to Windows.")
            args.target_os = "windows"
        elif "\\" in args.existing_path and "/" in args.new_path:
            logger.info("Auto detected target OS change. Will convert slashes to Linux/Mac.")
            args.target_os = "linux"
        else:
            args.target_os = None

    logger.debug(
        f"Existing Path: {args.existing_path}, New Path: {args.new_path}, "
        f"Target OS: {args.target_os}, Skip Bad Files: {args.skip_bad_files}"
    )
    qbm.run(args.existing_path, args.new_path, args.regex, args.target_os, True, args.skip_bad_files)


if __name__ == "__main__":
    main()  # pragma: no cover



import logging
import os
import re
from datetime import datetime
from threading import Thread
from typing import Optional

import bencodepy
from bencodepy.exceptions import BencodeDecodeError

from .methods import backup_folder, convert_slashes, discover_bt_backup_path


logger = logging.getLogger(__name__)
bencode = bencodepy.Bencode(encoding="utf-8", encoding_fallback="all")


class QBTBatchMove(object):
    logger = logging.getLogger(__name__ + ".QBTBatchMove")

    def __init__(self, bt_backup_path: str = None):
        if bt_backup_path is None:
            bt_backup_path = discover_bt_backup_path()
        self.logger.debug(f"BT_backup Path: {bt_backup_path}")
        self.bt_backup_path = bt_backup_path
        self.discovered_files = set()

    def run(
        self,
        existing_path: str,
        new_path: str,
        regex_path: bool = False,
        target_os: Optional[str] = None,
        create_backup: bool = True,
        skip_bad_files: bool = False,
    ):
        """
        Perform Batch Processing of path changes.
        :param existing_path: Existing path to look for
        :type existing_path: str
        :param new_path: New Path to replace with
        :type new_path: str
        :param regex_path: Existing and New Paths are regex patterns with capture groups
        :type regex_path: bool
        :param target_os: If targeting a different OS than the source. Must be Windows, Linux, or Mac.
        :type target_os: str
        :param create_backup: Create a backup archive of the BT_backup directory?
        :type create_backup: bool
        :param skip_bad_files: Skip .fastresume files that cannot be read successfully.
        :type skip_bad_files: bool
        """
        if not os.path.exists(self.bt_backup_path) or not os.path.isdir(self.bt_backup_path):
            raise NotADirectoryError(self.bt_backup_path)
        if create_backup:
            backup_filename = f'fastresume_backup{datetime.now().strftime("%Y%m%d%H%M%S")}.zip'
            backup_folder(self.bt_backup_path, os.path.join(self.bt_backup_path, backup_filename))

        self.logger.info(f"Searching for .fastresume files with path {existing_path} ...")
        for fast_resume in self.discover_relevant_fast_resume(
            self.bt_backup_path, existing_path, regex_path, not skip_bad_files
        ):
            # Fire and forget
            self.discovered_files.add(fast_resume)
            Thread(
                target=fast_resume.replace_paths, args=[existing_path, new_path, regex_path, target_os, True, False]
            ).start()

    @classmethod
    def discover_relevant_fast_resume(
        cls, bt_backup_path: str, existing_path: str, regex_path: bool = False, raise_on_error: bool = True
    ):
        """
        Find .fastresume files that contain the existing path.
        :param bt_backup_path: Path to BT_backup folder
        :type bt_backup_path: str
        :param existing_path: The existing path to look for
        :type existing_path: str
        :param regex_path: Existing Path is a regex pattern with capture groups
        :type: bool
        :param raise_on_error: Raise if error parsing .fastresume files
        :type raise_on_error: bool
        :return: List of FastResume Objects
        :rtype: list[FastResume]
        """
        for file in os.listdir(bt_backup_path):
            if file.endswith(".fastresume"):
                try:
                    fast_resume = FastResume(os.path.join(bt_backup_path, file))
                except (BencodeDecodeError, FileNotFoundError, ValueError) as e:
                    if raise_on_error:
                        cls.logger.critical(f"Unable to parse {file}. Stopping Discovery!")
                        raise e
                    cls.logger.warning(f"Unable to parse {file}. Skipping!\n\n{e}")
                    continue
                if existing_path in fast_resume.save_path or existing_path in fast_resume.qbt_save_path:
                    yield fast_resume
                elif regex_path and (
                    re.match(existing_path, fast_resume.save_path) or re.match(existing_path, fast_resume.qbt_save_path)
                ):
                    yield fast_resume
        return

    @classmethod
    def backup_folder(cls, folder_path: str, archive_path: str):
        return backup_folder(folder_path, archive_path)

    @classmethod
    def update_fastresume(
        cls,
        fast_resume: "FastResume",
        existing_path: str,
        new_path: str,
        regex_path: bool = False,
        target_os: Optional[str] = None,
        save_file: bool = True,
        create_backup: bool = True,
    ):
        if not isinstance(fast_resume, FastResume):
            raise TypeError("Not a FastResume object, cannot replace paths!")
        fast_resume.replace_paths(existing_path, new_path, regex_path, target_os, save_file, create_backup)


class FastResume(object):
    logger = logging.getLogger(__name__ + ".FastResume")

    def __init__(self, file_path: str):
        self._file_path = os.path.realpath(file_path)
        if not os.path.exists(self.file_path) or not os.path.isfile(self.file_path):
            raise FileNotFoundError(self.file_path)
        self.logger.debug(f"Loading Fast Resume: {self.file_path}")
        self._data = bencode.read(self.file_path)
        self.logger.debug(f"Fast Resume ({self.file_path}) Init Complete.")

    @property
    def file_path(self) -> str:
        return self._file_path

    @property
    def backup_filename(self) -> str:
        return f'{self.file_path}.{datetime.now().strftime("%Y%m%d%H%M%S")}.bkup'

    @property
    def save_path(self) -> Optional[str]:
        if "save_path" in self._data:
            return self._data["save_path"]

    @property
    def qbt_save_path(self) -> Optional[str]:
        if "qBt-savePath" in self._data:
            return self._data["qBt-savePath"]

    @property
    def mapped_files(self):
        if "mapped_files" in self._data:
            return self._data["mapped_files"]

    def set_save_path(
        self,
        path: str,
        key: str = "save_path",
        target_os: Optional[str] = None,
        save_file: bool = True,
        create_backup: bool = True,
    ):
        if key not in ["save_path", "qBt-savePath"]:
            raise KeyError("When setting a save path, key must be `save_path` or `qBt-savePath`. " f"Received {key}")
        if create_backup:
            self.save(self.backup_filename)
        if target_os is not None:
            path = convert_slashes(path, target_os)
        self.logger.debug(f"Setting {key}... Old: {self._data.get(key, None)}, New: {path}, Target OS: {target_os}")
        self._data[key] = path
        if save_file:
            self.save()

    def set_save_paths(
        self,
        path: str,
        qbt_path: Optional[str] = None,
        target_os: Optional[str] = None,
        save_file: bool = True,
        create_backup: bool = True,
    ):
        if create_backup:
            self.save(self.backup_filename)
        if not path.strip():
            raise ValueError("Cannot set empty paths!")
        self.set_save_path(path, key="save_path", target_os=target_os, save_file=False, create_backup=False)
        qbt_path = path if qbt_path is None else qbt_path
        if qbt_path:
            self.set_save_path(qbt_path, key="qBt-savePath", target_os=target_os, save_file=False, create_backup=False)
        if self.mapped_files is not None and target_os is not None:
            self.logger.debug("Converting Slashes for mapped_files...")
            self._data["mapped_files"] = [convert_slashes(path, target_os) for path in self.mapped_files]
        if save_file:
            self.save()

    def save(self, file_name: Optional[str] = None):
        if file_name is None:
            file_name = self.file_path
        self.logger.info(f"Saving File {file_name}...")
        bencode.write(self._data, file_name)

    def replace_paths(
        self,
        existing_path: str,
        new_path: str,
        regex_path: bool = False,
        target_os: Optional[str] = None,
        save_file: bool = True,
        create_backup: bool = True,
    ):
        self.logger.info(f"Replacing Paths in FastResume {self.file_path}...")
        if regex_path:
            new_save_path = None
            new_qbt_save_path = None
            pattern = re.compile(existing_path)
            if self.save_path:
                new_save_path = pattern.sub(new_path, self.save_path)
            if self.qbt_save_path:
                new_qbt_save_path = pattern.sub(new_path, self.qbt_save_path)
            if self.mapped_files:
                self._data["mapped_files"] = [pattern.sub(new_path, path) for path in self.mapped_files]
        else:
            new_save_path = self.save_path.replace(existing_path, new_path) if self.save_path is not None else None
            new_qbt_save_path = (
                self.qbt_save_path.replace(existing_path, new_path) if self.qbt_save_path is not None else None
            )
            if self.mapped_files:
                self._data["mapped_files"] = [path.replace(existing_path, new_path) for path in self.mapped_files]
        self.logger.debug(
            f"Existing Save Path: {existing_path}, New Save Path: {new_path}, " f"Replaced Save Path: {new_save_path}"
        )
        self.set_save_paths(
            path=str(new_save_path),
            qbt_path=new_qbt_save_path,
            target_os=target_os,
            save_file=save_file,
            create_backup=create_backup,
        )
        self.logger.info(f"FastResume ({self.file_path}) Paths Replaced!")



import logging
import os
import sys
import zipfile
from pathlib import Path
from typing import Union


logger = logging.getLogger(__name__)


def backup_folder(
    folder_path: Union[str, os.PathLike], archive_path: Union[str, os.PathLike], include_torrents: bool = True
):
    logger.info(f"Creating Archive {archive_path} ...")
    folder_path = Path(folder_path)
    archive_path = Path(archive_path)
    with zipfile.ZipFile(archive_path, "w") as archive:
        for file in folder_path.iterdir():
            if file.name.endswith(".fastresume") or (include_torrents and file.name.endswith(".torrent")):
                logger.debug(f"Archiving {file} into {archive_path}...")
                archive.write(file)
    logger.info("Done!")


def convert_slashes(path: str, target_os: str):
    if target_os.strip().lower() not in ("windows", "linux", "mac"):
        raise ValueError(f"Target OS is not valid. Must be Windows, Linux, or Mac. Received: {target_os}")
    if target_os.strip().lower() == "windows":
        logger.debug("Convert to Windows Slashes")
        return path.replace("/", "\\")
    logger.debug("Convert to Unix Slashes")
    return path.replace("\\", "/")


def discover_bt_backup_path():
    logger.debug("Discovering BT_backup path...")
    if sys.platform.startswith("win32"):
        logger.debug("Windows System")
        return os.path.join(os.getenv("localappdata"), "qBittorrent\\BT_backup")

    if os.path.isfile("/.dockerenv") and os.path.isdir("/config/qBittorrent/BT_backup"):
        # Default path for config under for image: https://docs.linuxserver.io/images/docker-qbittorrent
        logger.debug("qBittorrent Docker container detected")
        return "/config/qBittorrent/BT_backup"

    logger.debug("Linux/Mac System")
    return os.path.join(os.getenv("HOME", "/home"), ".local/share/data/qBittorrent/BT_backup")



import pytest

from qbt_migrate.cli import main, parse_args


class MockQBTBatchMove:
    instance = None
    run_call = None

    def __init__(self, bt_backup_path: str = None):
        self.bt_backup_path = bt_backup_path
        self.add_instance(self)

    @classmethod
    def run(cls, *args, **kwargs):
        cls.run_call = (args, kwargs)

    @classmethod
    def add_instance(cls, instance):
        cls.instance = instance


class MockUserInput:
    def __init__(self, items):
        self.index = 0
        self.items = items

    def next(self):
        if self.index < len(self.items):
            cur, self.index = self.items[self.index], self.index + 1
            print(f"Returning {cur}")
            return cur


def test_parse_args_defaults():
    # Check defaults
    args = parse_args([])
    assert all(
        hasattr(args, key)
        for key in ["existing_path", "new_path", "regex", "target_os", "bt_backup_path", "skip_bad_files", "log_level"]
    )
    assert args.existing_path is None
    assert args.new_path is None
    assert args.regex is None
    assert args.target_os is None
    assert args.bt_backup_path is None
    assert args.skip_bad_files is False
    assert args.log_level == "INFO"


def test_parse_args_shorthand():
    # Check shorthand values
    args = parse_args(
        ["-e", "existing-path", "-n", "new-path", "-r", "-t", "Linux", "-b", "bt-backup-path", "-s", "-l", "DEBUG"]
    )
    assert args.existing_path == "existing-path"
    assert args.new_path == "new-path"
    assert args.regex is True
    assert args.target_os == "Linux"
    assert args.bt_backup_path == "bt-backup-path"
    assert args.skip_bad_files is True
    assert args.log_level == "DEBUG"


def test_parse_args_longhand():
    # Check valid values
    args = parse_args(
        [
            "--existing-path",
            "existing-path",
            "--new-path",
            "new-path",
            "--regex",
            "--target-os",
            "Linux",
            "--bt-backup-path",
            "bt-backup-path",
            "--skip-bad-files",
            "--log-level",
            "DEBUG",
        ]
    )
    assert args.existing_path == "existing-path"
    assert args.new_path == "new-path"
    assert args.regex is True
    assert args.target_os == "Linux"
    assert args.bt_backup_path == "bt-backup-path"
    assert args.skip_bad_files is True
    assert args.log_level == "DEBUG"


def test_parse_args_target_os():
    args = parse_args(["-t", "Windows"])
    assert args.target_os == "Windows"
    args = parse_args(["-t", "Linux"])
    assert args.target_os == "Linux"
    args = parse_args(["-t", "Mac"])
    assert args.target_os == "Mac"

    with pytest.raises(SystemExit):
        args = parse_args(["-t", "windows"])


def test_main_with_inputs(monkeypatch):
    mock_user_input = MockUserInput(["bt-backup-path", "existing-path", "new-path", "no", "windows"])
    monkeypatch.setattr("builtins.input", lambda _: mock_user_input.next())
    monkeypatch.setattr("qbt_migrate.cli.QBTBatchMove", MockQBTBatchMove)
    monkeypatch.setattr("sys.argv", ["qbt_migrate"])
    main()
    assert MockQBTBatchMove.instance.bt_backup_path == "bt-backup-path"
    assert len(MockQBTBatchMove.run_call[0]) == 6
    assert MockQBTBatchMove.run_call[0][0] == "existing-path"
    assert MockQBTBatchMove.run_call[0][1] == "new-path"
    assert MockQBTBatchMove.run_call[0][2] is False
    assert MockQBTBatchMove.run_call[0][3] == "windows"
    assert MockQBTBatchMove.run_call[0][4] is True
    assert MockQBTBatchMove.run_call[0][5] is False


def test_main_with_args(monkeypatch):
    monkeypatch.setattr("qbt_migrate.cli.QBTBatchMove", MockQBTBatchMove)
    monkeypatch.setattr(
        "sys.argv",
        [
            "qbt_migrate",
            "-b",
            "different-bt-backup-path",
            "-e",
            "different-existing-path",
            "-n",
            "different-new-path",
            "-r",
            "-t",
            "Windows",
            "-s",
        ],
    )
    main()
    assert MockQBTBatchMove.instance.bt_backup_path == "different-bt-backup-path"
    assert len(MockQBTBatchMove.run_call[0]) == 6
    assert MockQBTBatchMove.run_call[0][0] == "different-existing-path"
    assert MockQBTBatchMove.run_call[0][1] == "different-new-path"
    assert MockQBTBatchMove.run_call[0][2] is True
    assert MockQBTBatchMove.run_call[0][3] == "Windows"
    assert MockQBTBatchMove.run_call[0][4] is True
    assert MockQBTBatchMove.run_call[0][5] is True


def test_main_invalid_input_loops(monkeypatch):
    monkeypatch.setattr("qbt_migrate.cli.QBTBatchMove", MockQBTBatchMove)
    monkeypatch.setattr("sys.argv", ["qbt_migrate", "-b", "backup-path", "-e", "e-path", "-n", "n-path", "-s"])
    mock_user_input = MockUserInput(["not-valid", "yes", "not-valid", "windows"])
    monkeypatch.setattr("builtins.input", lambda _: mock_user_input.next())
    main()
    assert MockQBTBatchMove.instance.bt_backup_path == "backup-path"
    assert len(MockQBTBatchMove.run_call[0]) == 6
    assert MockQBTBatchMove.run_call[0][0] == "e-path"
    assert MockQBTBatchMove.run_call[0][1] == "n-path"
    assert MockQBTBatchMove.run_call[0][2] is True
    assert MockQBTBatchMove.run_call[0][3] == "windows"
    assert MockQBTBatchMove.run_call[0][4] is True
    assert MockQBTBatchMove.run_call[0][5] is True


def test_main_target_os_auto_detect(monkeypatch):
    monkeypatch.setattr("qbt_migrate.cli.QBTBatchMove", MockQBTBatchMove)

    # Test no OS change
    mock_user_input = MockUserInput(["bt-backup-path", "existing-path", "new-path", "no", ""])
    monkeypatch.setattr("builtins.input", lambda _: mock_user_input.next())
    main()
    assert MockQBTBatchMove.run_call[0][3] is None

    # Test Windows to Linux
    mock_user_input = MockUserInput(["bt-backup-path", "C:\\existing\\path", "/new/path", "no", ""])
    monkeypatch.setattr("builtins.input", lambda _: mock_user_input.next())
    main()
    assert MockQBTBatchMove.run_call[0][3] == "linux"

    # Test Linux to Windows
    mock_user_input = MockUserInput(["bt-backup-path", "/existing/path", "C:\\new\\path", "no", ""])
    monkeypatch.setattr("builtins.input", lambda _: mock_user_input.next())
    main()
    assert MockQBTBatchMove.run_call[0][3] == "windows"



"""qBt Migrate, change the paths of existing torrents in qBittorrent, as well as convert paths to Windows/Linux/Mac"""
import logging
import os

from .classes import FastResume, QBTBatchMove
from .methods import convert_slashes, discover_bt_backup_path


__version__ = "2.2.0" + os.getenv("VERSION_TAG", "")

logging.getLogger(__name__).addHandler(logging.NullHandler())
